"""Git SCM plugin"""

from pathlib import Path

from cppython_core.plugin_schema.scm import SCM
from cppython_core.schema import Information
from dulwich.errors import NotGitRepository
from dulwich.repo import Repo


class GitSCM(SCM):
    """Git implementation hooks"""

    @staticmethod
    def supported(directory: Path) -> bool:
        """Queries repository status of a path

        Args:
            directory: The input path to query

        Returns:
            Whether the given path is a repository root
        """

        try:
            Repo(str(directory))
            return True

        except NotGitRepository:
            return False

    @staticmethod
    def information() -> Information:
        """Extracts the system's version metadata

        Returns:
            A version
        """
        return Information()

    def version(self, path: Path) -> str:
        """Extracts the system's version metadata

        Args:
            path: The repository path

        Returns:
            The git version
        """
        return ""

    def description(self) -> str | None:
        """Requests extraction of the project description"""
        return None



"""Tests the Project type"""


from pathlib import Path

import tomlkit
from cppython_core.schema import ProjectConfiguration
from pytest import FixtureRequest
from pytest_cppython.mock.interface import MockInterface

from cppython.project import Project


class TestProject:
    """Various tests for the project object"""

    def test_default_construction(self, request: FixtureRequest) -> None:
        """The project type should be constructable without pyproject.toml support.
        The CPPython project uses a working pyproject.toml file, and this file is used as the test data

        Args:
            request: The pytest request fixture
        """

        # Use the CPPython directory as the test data
        file = request.config.rootpath / "pyproject.toml"
        project_configuration = ProjectConfiguration(pyproject_file=file, version=None)
        interface = MockInterface()

        pyproject_data = tomlkit.loads(file.read_text(encoding="utf-8"))
        project = Project(project_configuration, interface, pyproject_data)

        assert project

    def test_missing_project_table(self, tmp_path: Path) -> None:
        """The project type should be constructable without the top level table

        Args:
            tmp_path: Temporary directory for dummy data
        """

        file_path = tmp_path / "pyproject.toml"

        with open(file_path, "a", encoding="utf8") as file:
            file.write("")

        project_configuration = ProjectConfiguration(pyproject_file=file_path, version=None)
        interface = MockInterface()

        project = Project(project_configuration, interface, {})

        assert project



"""Project schema specifications
"""

from abc import abstractmethod


class API:
    """Project API specification"""

    @abstractmethod
    def install(self) -> None:
        """Installs project dependencies"""
        raise NotImplementedError()

    @abstractmethod
    def update(self) -> None:
        """Updates project dependencies"""

        raise NotImplementedError()



"""Tests the click interface type"""

from click.testing import CliRunner

from cppython.console.interface import cli


class TestInterface:
    """Various tests for the click interface"""

    def test_info(self, cli_runner: CliRunner) -> None:
        """Verifies that the info command functions with CPPython hooks

        Args:
            cli_runner: The click runner
        """

        result = cli_runner.invoke(cli, ["info"], catch_exceptions=False)
        assert result.exit_code == 0

    def test_list(self, cli_runner: CliRunner) -> None:
        """Verifies that the list command functions with CPPython hooks

        Args:
            cli_runner: The click runner
        """

        result = cli_runner.invoke(cli, ["list"], catch_exceptions=False)
        assert result.exit_code == 0

    def test_update(self, cli_runner: CliRunner) -> None:
        """Verifies that the update command functions with CPPython hooks

        Args:
            cli_runner: The click runner
        """

        result = cli_runner.invoke(cli, ["update"], catch_exceptions=False)
        assert result.exit_code == 0

    def test_install(self, cli_runner: CliRunner) -> None:
        """Verifies that the install command functions with CPPython hooks

        Args:
            cli_runner: The click runner
        """

        result = cli_runner.invoke(cli, ["install"], catch_exceptions=False)
        assert result.exit_code == 0



"""Everything needed to build a CPPython project
"""

from importlib import metadata
from inspect import getmodule
from logging import Logger
from pathlib import Path
from typing import Any

from cppython_core.exceptions import ConfigError, PluginError
from cppython_core.plugin_schema.generator import Generator
from cppython_core.plugin_schema.provider import Provider
from cppython_core.plugin_schema.scm import SCM
from cppython_core.resolution import (
    resolve_cppython,
    resolve_cppython_plugin,
    resolve_generator,
    resolve_name,
    resolve_pep621,
    resolve_project_configuration,
    resolve_provider,
)
from cppython_core.schema import (
    CoreData,
    CorePluginData,
    CPPythonGlobalConfiguration,
    CPPythonLocalConfiguration,
    DataPluginT,
    PEP621Configuration,
    ProjectConfiguration,
)


class Builder:
    """Helper class for building CPPython projects"""

    def __init__(self, logger: Logger) -> None:
        self.logger = logger

    def generate_core_data(
        self,
        configuration: ProjectConfiguration,
        pep621_configuration: PEP621Configuration,
        cppython_configuration: CPPythonLocalConfiguration,
    ) -> CoreData:
        """Parses and returns resolved data from all configuration sources

        Args:
            configuration: Input configuration
            pep621_configuration: Project table configuration
            cppython_configuration: Tool configuration

        Raises:
            ConfigError: Raised if data cannot be parsed

        Returns:
            The resolved core object
        """

        global_configuration = CPPythonGlobalConfiguration()

        project_data = resolve_project_configuration(configuration)

        try:
            pep621_data = resolve_pep621(pep621_configuration, configuration)

        except ConfigError:
            configuration.version = self.extract_scm_version(configuration.pyproject_file.parent)
            pep621_data = resolve_pep621(pep621_configuration, configuration)

        cppython_data = resolve_cppython(cppython_configuration, global_configuration, project_data)

        return CoreData(project_data=project_data, pep621_data=pep621_data, cppython_data=cppython_data)

    def extract_scm_version(self, path: Path) -> str:
        """Locates an available SCM plugin that can report version information about the given path

        Args:
            path: The directory to query

        Raises:
            PluginError: If no SCM plugin can be found

        Returns:
            A version token
        """

        group = "SCM"
        group_lower = group.lower()

        scm_types: list[type[SCM]] = []

        if not (entries := list(metadata.entry_points(group=f"cppython.{group_lower}"))):
            raise PluginError("No SCM plugin found")

        # Filter entries
        for entry_point in entries:
            plugin_type = entry_point.load()
            if not issubclass(plugin_type, SCM):
                self.logger.warning(
                    f"Found incompatible plugin. The '{resolve_name(plugin_type)}' plugin must be an instance of"
                    f" '{group_lower}'"
                )
            else:
                scm_types.append(plugin_type)

        # Deduce the SCM repository
        plugin = None
        for scm_type in scm_types:
            scm = scm_type()
            if scm.supported(path):
                plugin = scm
                break

        if not plugin:
            raise PluginError("No applicable SCM plugin found for the given path")

        if (version := plugin.version(path)) is None:
            raise PluginError("Project has no version information")

        return version

    def find_generators(self) -> list[type[Generator]]:
        """_summary_

        Raises:
            PluginError: _description_

        Returns:
            _description_
        """

        group_name = "generator"
        plugin_types: list[type[Generator]] = []

        # Filter entries by type
        for entry_point in list(metadata.entry_points(group=f"cppython.{group_name}")):
            loaded_type = entry_point.load()
            if not issubclass(loaded_type, Generator):
                self.logger.warning(
                    f"Found incompatible plugin. The '{resolve_name(loaded_type)}' plugin must be an instance of"
                    f" '{group_name}'"
                )
            else:
                self.logger.warning(
                    f"{group_name} plugin found: {resolve_name(loaded_type)} from {getmodule(loaded_type)}"
                )
                plugin_types.append(loaded_type)

        if not plugin_types:
            raise PluginError(f"No {group_name} plugin was found")

        return plugin_types

    def find_providers(self) -> list[type[Provider]]:
        """_summary_

        Raises:
            PluginError: _description_

        Returns:
            _description_
        """

        group_name = "provider"
        plugin_types: list[type[Provider]] = []

        # Filter entries by type
        for entry_point in list(metadata.entry_points(group=f"cppython.{group_name}")):
            loaded_type = entry_point.load()
            if not issubclass(loaded_type, Provider):
                self.logger.warning(
                    f"Found incompatible plugin. The '{resolve_name(loaded_type)}' plugin must be an instance of"
                    f" '{group_name}'"
                )
            else:
                self.logger.warning(
                    f"{group_name} plugin found: {resolve_name(loaded_type)} from {getmodule(loaded_type)}"
                )
                plugin_types.append(loaded_type)

        if not plugin_types:
            raise PluginError(f"No {group_name} plugin was found")

        return plugin_types

    def filter_plugins(
        self, plugin_types: list[type[DataPluginT]], directory: Path, pinned_name: str | None, group_name: str
    ) -> list[type[DataPluginT]]:
        """Finds and filters data plugins

        Args:
            plugin_types: The plugin type to lookup
            directory: The data to query support for the filtered plugins
            pinned_name: The configuration name
            group_name: The group name

        Raises:
            PluginError: Raised if no plugins can be found

        Returns:
            The list of applicable plugins
        """

        # Lookup the requested plugin if given
        if pinned_name is not None:
            for loaded_type in plugin_types:
                if resolve_name(loaded_type) == pinned_name:
                    self.logger.warning(
                        f"Using {group_name} plugin: {resolve_name(loaded_type)} from {getmodule(loaded_type)}"
                    )
                    return [loaded_type]

        self.logger.warning(f"'{group_name}_name' was empty. Trying to deduce {group_name}s")

        supported_types: list[type[DataPluginT]] = []

        # Deduce types
        for loaded_type in plugin_types:
            if loaded_type.supported(directory):
                self.logger.warning(
                    f"A {group_name} plugin is supported: {resolve_name(loaded_type)} from {getmodule(loaded_type)}"
                )
                supported_types.append(loaded_type)

        # Fail
        if supported_types is None:
            raise PluginError(f"No {group_name} could be deduced from the root directory.")

        return supported_types

    def solve(
        self, generator_types: list[type[Generator]], provider_types: list[type[Provider]]
    ) -> tuple[type[Generator], type[Provider]]:
        """_summary_

        Args:
            generator_types: _description_
            provider_types: _description_

        Raises:
            PluginError: _description_

        Returns:
            _description_
        """

        combos: list[tuple[type[Generator], type[Provider]]] = []

        for generator_type in generator_types:
            sync_types = generator_type.sync_types()
            for provider_type in provider_types:
                for sync_type in sync_types:
                    if provider_type.supported_sync_type(sync_type):
                        combos.append((generator_type, provider_type))
                        break

        if not combos:
            raise PluginError("No provider that supports a given generator could be deduced")

        return combos[0]

    def create_generator(
        self, core_data: CoreData, generator_configuration: dict[str, Any], generator_type: type[Generator]
    ) -> Generator:
        """Creates a generator from input configuration

        Args:
            core_data: The resolved configuration data
            generator_configuration: The generator table of the CPPython configuration data
            generator_type: The plugin type

        Raises:
            PluginError: Raised if no viable generator plugin was found

        Returns:
            The constructed generator
        """

        generator_data = resolve_generator(core_data.project_data)
        cppython_plugin_data = resolve_cppython_plugin(core_data.cppython_data, generator_type)

        core_plugin_data = CorePluginData(
            project_data=core_data.project_data,
            pep621_data=core_data.pep621_data,
            cppython_data=cppython_plugin_data,
        )

        if not generator_configuration:
            self.logger.error(
                "The pyproject.toml table 'tool.cppython.generator' does not exist. Sending generator empty data",
            )

        return generator_type(generator_data, core_plugin_data, generator_configuration)

    def create_provider(
        self, core_data: CoreData, provider_configuration: dict[str, Any], provider_type: type[Provider]
    ) -> Provider:
        """Creates Providers from input data

        Args:
            core_data: The resolved configuration data
            provider_configuration: The provider data table
            provider_type: The type to instantiate

        Raises:
            PluginError: Raised if no viable provider plugin was found

        Returns:
            A constructed provider plugins
        """

        provider_data = resolve_provider(core_data.project_data)
        cppython_plugin_data = resolve_cppython_plugin(core_data.cppython_data, provider_type)

        core_plugin_data = CorePluginData(
            project_data=core_data.project_data,
            pep621_data=core_data.pep621_data,
            cppython_data=cppython_plugin_data,
        )

        if not provider_configuration:
            self.logger.error(
                "The pyproject.toml table 'tool.cppython.provider' does not exist. Sending provider empty data",
            )

        return provider_type(provider_data, core_plugin_data, provider_configuration)



"""Manages data flow to and from plugins
"""

import asyncio
import logging
from typing import Any

from cppython_core.exceptions import ConfigError, PluginError
from cppython_core.resolution import resolve_name
from cppython_core.schema import CoreData, Interface, ProjectConfiguration, PyProject
from pydantic import ValidationError

from cppython.builder import Builder
from cppython.schema import API


class Project(API):
    """The object constructed at each entry_point"""

    def __init__(
        self, configuration: ProjectConfiguration, interface: Interface, pyproject_data: dict[str, Any]
    ) -> None:
        self._enabled = False
        self._interface = interface

        # Default logging levels
        levels = [logging.WARNING, logging.INFO, logging.DEBUG]

        # Add default output stream
        self.logger = logging.getLogger("cppython")
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(levels[configuration.verbosity])

        self.logger.info("Initializing project")

        try:
            if (pyproject := PyProject(**pyproject_data)) is None:
                raise ConfigError("Table [project] is not defined")

            if pyproject.tool is None:
                raise ConfigError("Table [tool] is not defined")

            if pyproject.tool.cppython is None:
                raise ConfigError("Table [tool.cppython] is not defined")

            builder = Builder(self.logger)

            raw_generator_plugins = builder.find_generators()
            generator_plugins = builder.filter_plugins(
                raw_generator_plugins,
                configuration.pyproject_file.parent,
                pyproject.tool.cppython.generator_name,
                "Generator",
            )

            raw_provider_plugins = builder.find_providers()
            provider_plugins = builder.filter_plugins(
                raw_provider_plugins,
                configuration.pyproject_file.parent,
                pyproject.tool.cppython.provider_name,
                "Provider",
            )

            # Solve the messy interactions between plugins
            generator_type, provider_type = builder.solve(generator_plugins, provider_plugins)

            # Once the plugins are resolved, the core data is complete and can be generated
            self._core_data = builder.generate_core_data(configuration, pyproject.project, pyproject.tool.cppython)

            # Create the chosen plugins
            self._generator = builder.create_generator(
                self.core_data, pyproject.tool.cppython.generator, generator_type
            )
            self._provider = builder.create_provider(self.core_data, pyproject.tool.cppython.provider, provider_type)

        except ConfigError:
            logging.exception("Unhandled configuration. CPPython will process no further")
            return
        except ValidationError as error:
            logging.error(error)
            return

        self._enabled = True

        self.logger.info("Initialized project successfully")

    @property
    def enabled(self) -> bool:
        """Queries if the project was is initialized for full functionality

        Returns:
            The query result
        """
        return self._enabled

    @property
    def core_data(self) -> CoreData:
        """Core data

        Returns:
            Core data
        """
        return self._core_data

    async def download_provider_tools(self) -> None:
        """Download the provider tooling if required"""
        if not self._enabled:
            self.logger.info("Skipping 'download_provider_tools' because the project is not enabled")
            return

        name = resolve_name(type(self._provider))
        base_path = self.core_data.cppython_data.install_path

        path = base_path / name

        path.mkdir(parents=True, exist_ok=True)

        self.logger.warning("Downloading the %s requirements to %s", name, path)
        await self._provider.download_tooling(path)

    def sync(self) -> None:
        """Gathers sync information from providers and passes it to the generator

        Raises:
            PluginError: Plugin error
        """

        if (sync_data := self._provider.sync_data(self._generator)) is None:
            raise PluginError("The provider doesn't support the generator")

        self._generator.sync(sync_data)

    # API Contract
    def install(self) -> None:
        """Installs project dependencies

        Raises:
            Exception: Raised if failed
        """
        if not self._enabled:
            self.logger.info("Skipping install because the project is not enabled")
            return

        self.logger.info("Installing tools")
        asyncio.run(self.download_provider_tools())

        self.logger.info("Installing project")
        name = resolve_name(type(self._provider))
        self.logger.info("Installing %s provider", name)

        try:
            self._provider.install()
        except Exception as exception:
            self.logger.error("Provider %s failed to install", name)
            raise exception

        self.sync()

    def update(self) -> None:
        """Updates project dependencies

        Raises:
            Exception: Raised if failed
        """
        if not self._enabled:
            self.logger.info("Skipping update because the project is not enabled")
            return

        self.logger.info("Updating tools")
        asyncio.run(self.download_provider_tools())

        self.logger.info("Updating project")
        name = resolve_name(type(self._provider))
        self.logger.info("Updating %s provider", name)

        try:
            self._provider.update()
        except Exception as exception:
            self.logger.error("Provider %s failed to update", name)
            raise exception

        self.sync()



"""Tests the cppython built-in SCM plugin
"""

import pytest
from pytest_cppython.tests import SCMIntegrationTests

from cppython.plugins.git import GitSCM


class TestGitInterface(SCMIntegrationTests[GitSCM]):
    """Integration tests for the Git SCM plugin"""

    @pytest.fixture(name="plugin_type", scope="session")
    def fixture_plugin_type(self) -> type[GitSCM]:
        """A required testing hook that allows type generation

        Returns:
            The SCM type
        """
        return GitSCM
